@@ -1,5 +1,5 @@
 """Functionality for autorenewal and associated juggling of configurations"""
-
+import configobj
 import copy
 import datetime
 import itertools
@@ -9,9 +9,7 @@ import sys
 import time
 import traceback
 from typing import Any
-from typing import Dict
 from typing import Iterable
-from typing import List
 from typing import Mapping
 from typing import Optional
 from typing import Union
@@ -32,6 +30,7 @@ from certbot._internal import cli
 from certbot._internal import client
 from certbot._internal import constants
 from certbot._internal import hooks
+from certbot._internal import san
 from certbot._internal import storage
 from certbot._internal import updater
 from certbot._internal.display import obj as display_obj
@@ -41,6 +40,8 @@ from certbot.display import util as display_util
 
 logger = logging.getLogger(__name__)
 
+ARI_RETRY_AFTER_CONFIG_ITEM = "ari_retry_after"
+
 # These are the items which get pulled out of a renewal configuration
 # file's renewalparams and actually used in the client configuration
 # during the renewal process. We have to record their types here because
@@ -57,6 +58,39 @@ BOOL_CONFIG_ITEMS = ["must_staple", "allow_subset_of_names", "reuse_key",
 CONFIG_ITEMS = set(itertools.chain(
     BOOL_CONFIG_ITEMS, INT_CONFIG_ITEMS, STR_CONFIG_ITEMS, ('pref_challs',)))
 
+class AriClientPool:
+    """A cache of ACME clients for using in performing ACME Renewal Info (ARI) requests.
+
+    During `certbot renew` we need to check ARI for many certificates, and usually these
+    are all issued by the same server. To avoid redundant directory fetches, we create
+    one acme.ClientV2 per server.
+
+    This takes a command line configuration object so it can set the User-Agent header and
+    observe the --no-verify-ssl flag.
+    """
+    def __init__(self, cli_config: configuration.NamespaceConfig):
+        self._verify_ssl = not cli_config.no_verify_ssl
+        self._user_agent = client.determine_user_agent(cli_config)
+        self._pool: dict[str, acme_client.ClientV2] = {}
+
+    def get(self, server: str) -> acme_client.ClientV2:
+        """
+        Retrieve or create an ACME client for the specified server.
+
+        Returns:
+            acme_client.ClientV2: The ACME client associated with the specified server.
+        """
+        ari_client = self._pool.get(server, None)
+        if ari_client:
+            return ari_client
+
+        net = acme_client.ClientNetwork(verify_ssl=self._verify_ssl, user_agent=self._user_agent)
+        directory = acme_client.ClientV2.get_directory(server, net)
+        ari_client = acme_client.ClientV2(directory, net)
+
+        self._pool[server] = ari_client
+        return ari_client
+
 
 def reconstitute(config: configuration.NamespaceConfig,
                   full_path: str) -> Optional[storage.RenewableCert]:
@@ -114,8 +148,8 @@ def reconstitute(config: configuration.NamespaceConfig,
         return None
 
     try:
-        config.domains = [util.enforce_domain_sanity(d)
-                          for d in renewal_candidate.names()]
+        domains, _ = san.split(renewal_candidate.sans())
+        config.domains = domains
     except errors.ConfigurationError as error:
         logger.error("Renewal configuration file %s references a certificate "
                        "that contains an invalid domain name. The problem "
@@ -163,7 +197,7 @@ def _restore_plugin_configs(config: configuration.NamespaceConfig,
     #      longer defined, stored copies of that parameter will be
     #      deserialized as strings by this logic even if they were
     #      originally meant to be some other type.
-    plugin_prefixes: List[str] = []
+    plugin_prefixes: list[str] = []
     if renewalparams["authenticator"] == "webroot":
         _restore_webroot_config(config, renewalparams)
     else:
@@ -212,7 +246,7 @@ def restore_required_config_elements(config: configuration.NamespaceConfig,
         setattr(config, key, value)
 
 
-def _remove_deprecated_config_elements(renewalparams: Mapping[str, Any]) -> Dict[str, Any]:
+def _remove_deprecated_config_elements(renewalparams: Mapping[str, Any]) -> dict[str, Any]:
     """Removes deprecated config options from the parsed renewalparams.
 
     :param dict renewalparams: list of parsed renewalparams
@@ -225,7 +259,7 @@ def _remove_deprecated_config_elements(renewalparams: Mapping[str, Any]) -> Dict
         if option_name not in cli.DEPRECATED_OPTIONS}
 
 
-def _restore_pref_challs(unused_name: str, value: Union[List[str], str]) -> List[str]:
+def _restore_pref_challs(unused_name: str, value: Union[list[str], str]) -> list[str]:
     """Restores preferred challenges from a renewal config file.
 
     If value is a `str`, it should be a single challenge type.
@@ -310,7 +344,7 @@ def _restore_str(name: str, value: str) -> Optional[str]:
 
 def should_renew(config: configuration.NamespaceConfig,
                  lineage: storage.RenewableCert,
-                 acme_clients: Dict[str, acme_client.ClientV2]) -> bool:
+                 ari_clients: AriClientPool) -> bool:
     """Return true if any of the circumstances for automatic renewal apply."""
     if config.renew_by_default:
         logger.debug("Auto-renewal forced with --force-renewal...")
@@ -318,41 +352,43 @@ def should_renew(config: configuration.NamespaceConfig,
     if config.dry_run:
         logger.info("Certificate not due for renewal, but simulating renewal for dry run")
         return True
-    if should_autorenew(config, lineage, acme_clients):
+    if should_autorenew(lineage, ari_clients):
         logger.info("Certificate is due for renewal, auto-renewing...")
         return True
     display_util.notify("Certificate not yet due for renewal")
     return False
 
 
-def _ari_renewal_time(config: configuration.NamespaceConfig,
-                      lineage: storage.RenewableCert,
-                      acme_clients: Dict[str, acme_client.ClientV2],
-                      cert_pem: bytes,
-                      )-> Optional[datetime.datetime]:
+def _ari_renewal_time(lineage: storage.RenewableCert,
+                     cert_pem: bytes,
+                     ari_clients: AriClientPool) -> Optional[datetime.datetime]:
     """Return the ARI suggested renewal time if it's available."""
     # For ARI requests, we want to use the ACME directory URL from which the
-    # cert was originally requested. Since `config.server` can be overridden on
+    # cert was originally requested. Since `NamespaceConfig.server` can be overridden on
     # the command line, we're using the server stored in the cert's renewal
     # conf, i.e. `lineage.server`
     #
     # Fixes https://github.com/certbot/certbot/issues/10339
     if not lineage.server:
-        renewal_conf_file = storage.renewal_filename_for_lineagename(config, lineage.lineagename)
         logger.warning("Skipping ARI check because %s has no 'server' field. This issue will not "
-                       "prevent certificate renewal", renewal_conf_file)
+                       "prevent certificate renewal", lineage.configfile.filename)
         return None
-    try:
-        # Creating a new ACME client makes a network request, so check if we have
-        # one cached for this cert's server already
-        if lineage.server not in acme_clients:
-            acme_clients[lineage.server] = \
-                client.create_acme_client(config, server_override=lineage.server)
-        acme = acme_clients.get(lineage.server, None)
 
-        # Attempt to get the ARI-defined renewal time
-        if acme:
-            return acme.renewal_time(cert_pem)[0]
+    ari_config_section = lineage.configfile.get("acme_renewal_info", {})
+    retry_after = ari_config_section.get(ARI_RETRY_AFTER_CONFIG_ITEM, None)
+    if retry_after:
+        retry_after_datetime = datetime.datetime.fromisoformat(retry_after)
+        now = datetime.datetime.now()
+        if now < retry_after_datetime:
+            logger.debug("Skipped ACME Renewal Info check because ari_retry_after %s is in "
+                         "the future",
+                         retry_after)
+            return None
+
+    renewal_time = None
+    try:
+        ari_client = ari_clients.get(lineage.server)
+        renewal_time, retry_after = ari_client.renewal_time(cert_pem)
     except Exception:  # pylint: disable=broad-except
         # We want to stop errors around ARI preventing renewal so we catch all exceptions here
         # with a warning asking users to tell us about any problems they are experiencing
@@ -360,8 +396,17 @@ def _ari_renewal_time(config: configuration.NamespaceConfig,
                        "problem persists and you think it's a bug in Certbot, please open an "
                        "issue at https://github.com/certbot/certbot/issues/new/choose.")
         logger.debug("Error while requesting ARI was:", exc_info=True)
+        retry_after = datetime.datetime.now() + datetime.timedelta(seconds=60 * 60 * 6)
 
-    return None
+
+    config_update = configobj.ConfigObj()
+    config_update["acme_renewal_info"] = {
+        # Note: the ACME client returns naive (no timezone) datetimes for retry_after, and that
+        # is what we serialize here.
+        ARI_RETRY_AFTER_CONFIG_ITEM: retry_after.isoformat(timespec="seconds"),
+    }
+    storage.atomic_rewrite(lineage.configfile.filename, config_update)
+    return renewal_time
 
 
 def _default_renewal_time(cert_pem: bytes) -> datetime.datetime:
@@ -384,9 +429,8 @@ def _default_renewal_time(cert_pem: bytes) -> datetime.datetime:
 
     return default_rt
 
-def should_autorenew(config: configuration.NamespaceConfig,
-                     lineage: storage.RenewableCert,
-                     acme_clients: Dict[str, acme_client.ClientV2]) -> bool:
+def should_autorenew(lineage: storage.RenewableCert,
+                     ari_clients: AriClientPool) -> bool:
     """Should we now try to autorenew the most recent cert version?
 
     If automatic renewal is disabled for the lineage, this function
@@ -414,7 +458,7 @@ def should_autorenew(config: configuration.NamespaceConfig,
     with open(cert, 'rb') as f:
         cert_pem = f.read()
 
-    renewal_time = _ari_renewal_time(config, lineage, acme_clients, cert_pem)
+    renewal_time = _ari_renewal_time(lineage, cert_pem, ari_clients)
 
     now = datetime.datetime.now(datetime.timezone.utc)
 
@@ -450,7 +494,7 @@ def _avoid_invalidating_lineage(config: configuration.NamespaceConfig,
     if util.is_staging(config.server):
         if not util.is_staging(original_server):
             if not config.break_my_certs:
-                names = ", ".join(lineage.names())
+                names = san.display(lineage.sans())
                 raise errors.Error(
                     "You've asked to renew/replace a seemingly valid certificate with "
                     f"a test certificate (domains: {names}). We will not do that "
@@ -501,15 +545,15 @@ def _avoid_reuse_key_conflicts(config: configuration.NamespaceConfig,
                 "add --new-key.")
 
 
-def renew_cert(config: configuration.NamespaceConfig, domains: Optional[List[str]],
+def renew_cert(config: configuration.NamespaceConfig, sans: Optional[list[san.SAN]],
                le_client: client.Client, lineage: storage.RenewableCert) -> None:
     """Renew a certificate lineage."""
     renewal_params = lineage.configuration["renewalparams"]
     original_server = renewal_params.get("server", cli.flag_default("server"))
     _avoid_invalidating_lineage(config, lineage, original_server)
     _avoid_reuse_key_conflicts(config, lineage)
-    if not domains:
-        domains = lineage.names()
+    if not sans:
+        sans = lineage.sans()
     # The private key is the existing lineage private key if reuse_key is set.
     # Otherwise, generate a fresh private key by passing None.
     if config.reuse_key and not config.new_key:
@@ -517,7 +561,7 @@ def renew_cert(config: configuration.NamespaceConfig, domains: Optional[List[str
         _update_renewal_params_from_key(new_key, config)
     else:
         new_key = None
-    new_cert, new_chain, new_key, _ = le_client.obtain_certificate(domains, new_key)
+    new_cert, new_chain, new_key, _ = le_client.obtain_certificate(sans, new_key)
     if config.dry_run:
         logger.debug("Dry run: skipping updating lineage at %s", os.path.dirname(lineage.cert))
     else:
@@ -527,7 +571,7 @@ def renew_cert(config: configuration.NamespaceConfig, domains: Optional[List[str
         lineage.update_all_links_to(lineage.latest_common_version())
         lineage.truncate()
 
-    hooks.renew_hook(config, domains, lineage.live_dir)
+    hooks.renew_hook(config, sans, lineage.live_dir)
 
 
 def report(msgs: Iterable[str], category: str) -> str:
@@ -536,9 +580,9 @@ def report(msgs: Iterable[str], category: str) -> str:
     return "  " + "\n  ".join(lines)
 
 
-def _renew_describe_results(config: configuration.NamespaceConfig, renew_successes: List[str],
-                            renew_failures: List[str], renew_skipped: List[str],
-                            parse_failures: List[str]) -> None:
+def _renew_describe_results(config: configuration.NamespaceConfig, renew_successes: list[str],
+                            renew_failures: list[str], renew_skipped: list[str],
+                            parse_failures: list[str]) -> None:
     """
     Print a report to the terminal about the results of the renewal process.
 
@@ -588,7 +632,7 @@ def handle_renewal_request(config: configuration.NamespaceConfig) -> None:
     """Examine each lineage; renew if due and report results"""
 
     # This is trivially False if config.domains is empty
-    if any(domain not in config.webroot_map for domain in config.domains):
+    if any(domain.dns_name not in config.webroot_map for domain in config.domains):
         # If more plugins start using cli.add_domains,
         # we may want to only log a warning here
         raise errors.Error("Currently, the renew verb is capable of either "
@@ -622,7 +666,7 @@ def handle_renewal_request(config: configuration.NamespaceConfig) -> None:
     # We initialize acme clients on a per-server basis, but most
     # lineages use the same server. Memoize clients here so we can
     # share the connection pool and reuse a single fetched directory.
-    acme_clients: Dict[str, acme_client.ClientV2] = {}
+    ari_clients = AriClientPool(config)
 
     for renewal_file in conf_files:
         display_util.notification("Processing " + renewal_file, pause=False)
@@ -649,7 +693,7 @@ def handle_renewal_request(config: configuration.NamespaceConfig) -> None:
                 renewal_candidate.ensure_deployed()
                 from certbot._internal import main
                 plugins = plugins_disco.PluginsRegistry.find_all()
-                if should_renew(lineage_config, renewal_candidate, acme_clients):
+                if should_renew(lineage_config, renewal_candidate, ari_clients):
                     # Apply random sleep upon first renewal if needed
                     if apply_random_sleep:
                         sleep_time = random.uniform(1, 60 * 8)
@@ -666,7 +710,7 @@ def handle_renewal_request(config: configuration.NamespaceConfig) -> None:
                     # and we have a lineage in renewal_candidate
                     main.renew_cert(lineage_config, plugins, renewal_candidate)
                     renew_successes.append(renewal_candidate.fullchain)
-                    renewed_domains.extend(renewal_candidate.names())
+                    renewed_domains.extend(renewal_candidate.sans())
                 else:
                     expiry = crypto_util.notAfter(renewal_candidate.version(
                         "cert", renewal_candidate.latest_common_version()))
@@ -677,7 +721,7 @@ def handle_renewal_request(config: configuration.NamespaceConfig) -> None:
                                              plugins)
 
         except Exception as e:  # pylint: disable=broad-except
-            # obtain_cert (presumably) encountered an unanticipated problem.
+            # obtain_certificate (presumably) encountered an unanticipated problem.
             logger.error(
                 "Failed to renew certificate %s with error: %s",
                 lineagename, e
@@ -685,7 +729,7 @@ def handle_renewal_request(config: configuration.NamespaceConfig) -> None:
             logger.debug("Traceback was:\n%s", traceback.format_exc())
             if renewal_candidate:
                 renew_failures.append(renewal_candidate.fullchain)
-                failed_domains.extend(renewal_candidate.names())
+                failed_domains.extend(renewal_candidate.sans())
 
     # Describe all the results
     _renew_describe_results(config, renew_successes, renew_failures,
